#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "ble_db_discovery.h"

#include "fstorage.h"
#include "fds.h"
#include "peer_manager.h"
#include "ble_conn_state.h"

#include "nordic_common.h"
#include "ble.h"
#include "ble_hci.h"
#include "softdevice_handler.h"
#include "app_util.h"
#include "app_error.h"
#include "boards.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "app_timer.h"

#include "uart.h"

#include "SEGGER_RTT.h"

#define DK_BOARD

#ifdef DK_BOARD
	#define RED_LED				17
	#define LED_ON(x)			nrf_gpio_pin_clear(x)	
	#define LED_OFF(x)		nrf_gpio_pin_set(x)  
	#define TAG_STATUS		19
	#define TAG_STATUS_ON		nrf_gpio_pin_clear(TAG_STATUS)
	#define TAG_STATUS_OFF	nrf_gpio_pin_set(TAG_STATUS)
#else
	#define RED_LED				1
	#define LED_ON(x)			nrf_gpio_pin_set(x)
	#define LED_OFF(x)		nrf_gpio_pin_clear(x)
	#define TAG_STATUS		7
	#define TAG_STATUS_ON		nrf_gpio_pin_clear(TAG_STATUS)
	#define TAG_STATUS_OFF	nrf_gpio_pin_set(TAG_STATUS)
#endif

#define MIN_TX_POWER_LEVEL				 0
#define MAX_TX_POWER_LEVEL				 0

#define CENTRAL_LINK_COUNT         1                                  /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT      0                                  /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

#define APP_TIMER_PRESCALER        0                                  /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE    2                                  /**< Size of timer operation queues. */

#define SCAN_INTERVAL              BLE_GAP_SCAN_INTERVAL_MAX          /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW                BLE_GAP_SCAN_WINDOW_MAX          	/**< Determines scan window in units of 0.625 millisecond. */

#define SEC_PARAM_BOND             1                                  /**< Perform bonding. */
#define SEC_PARAM_MITM             1                                  /**< Man In The Middle protection not required. */
#define SEC_PARAM_LESC             0                                  /**< LE Secure Connections not enabled. */
#define SEC_PARAM_KEYPRESS         0                                  /**< Keypress notifications not enabled. */
#define SEC_PARAM_IO_CAPABILITIES  BLE_GAP_IO_CAPS_DISPLAY_ONLY //BLE_GAP_IO_CAPS_NONE   						/**< No I/O capabilities. */
#define SEC_PARAM_OOB              0                                  /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE     7                                  /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE     16                                 /**< Maximum encryption key size. */

#define RED 1

typedef struct
{
    uint8_t     * p_data;                                             /**< Pointer to data. */
    uint16_t      data_len;                                           /**< Length of data. */
}data_t;

typedef enum
{
    BLE_NO_SCAN,                                                     	/**< No advertising running. */
    BLE_WHITELIST_SCAN,                                              	/**< Advertising with whitelist. */
    BLE_FAST_SCAN,                                                   	/**< Fast advertising running. */
} ble_scan_mode_t;

static const ble_gap_conn_params_t m_conn_fast_param =
{
    .min_conn_interval = MSEC_TO_UNITS(20, UNIT_1_25_MS),   
    .max_conn_interval = MSEC_TO_UNITS(120, UNIT_1_25_MS),   
    .slave_latency = 0,                                 
    .conn_sup_timeout = MSEC_TO_UNITS(8000, UNIT_10_MS)        
};

static const ble_gap_conn_params_t m_conn_slow_param =
{
    .min_conn_interval = MSEC_TO_UNITS(BLE_GAP_CP_MIN_CONN_INTVL_MAX, UNIT_1_25_MS),   
    .max_conn_interval = MSEC_TO_UNITS(BLE_GAP_CP_MIN_CONN_INTVL_MAX, UNIT_1_25_MS),   
    .slave_latency = 0,                                 
    .conn_sup_timeout = MSEC_TO_UNITS(8000, UNIT_10_MS)        
};

ble_db_discovery_t           				m_ble_db_discovery;               /**< Structure used to identify the DB Discovery module. */

static ble_scan_mode_t              m_scan_mode = BLE_WHITELIST_SCAN; /**< Scan mode used by application. */
static ble_gap_scan_params_t        m_scan_param;                     /**< Scan parameters requested for scanning and connection. */
static ble_gap_whitelist_t   				m_whitelist;
static ble_gap_addr_t        			* p_whitelist_addr[BLE_GAP_WHITELIST_ADDR_MAX_COUNT];
static ble_gap_irk_t         			* p_whitelist_irk[BLE_GAP_WHITELIST_IRK_MAX_COUNT];

static volatile bool                m_whitelist_temporarily_disabled 	= false; /**< True if whitelist has been temporarily disabled. */
static bool                         m_memory_access_in_progress 			= false; /**< Flag to keep track of ongoing operations on persistent memory. */

static uint32_t 										m_start_cnt;
static uint32_t 										m_stop_cnt;
static uint32_t 										m_diff_cnt;

#pragma push
#pragma pack(1)
#pragma anon_unions

#define MAX_STRING_SIZE	16

typedef __packed struct
{
	uint16_t	major;
	uint16_t	minor;
} VERSION;

typedef struct
{
	uint8_t			revision;
	uint8_t 		name_lenght;
	const char 	name[MAX_STRING_SIZE];
	uint8_t 		serial_number_lenght;
	const char 	serial_number[MAX_STRING_SIZE];
	uint8_t 		passkey_length;
	const char 	passkey[MAX_STRING_SIZE];
	VERSION 		version;	
	const char	data_and_time[18];
	VERSION 		min_pc_version;
} HEX_CONTEXT;

#pragma pop

const HEX_CONTEXT hex __attribute__((section(".ARM.__at_0x1F800"))) = 
			{
				.revision							= 0,
				.name_lenght 					= 6, .name 					= "Igla_2_", 
				.serial_number_lenght = 6, .serial_number = "000001", 
				.passkey_length 			= 6, .passkey 			= "123456", 
				.version.major 				= 2, .version.minor = 0, 
				.data_and_time 				= "20151029T120000+03", 
				.min_pc_version.major = 2, .min_pc_version.minor = 0
			};
			

static void scan_start(void);

void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}

static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
	uint32_t  index = 0;
	uint8_t * p_data;

	p_data = p_advdata->p_data;

	while (index < p_advdata->data_len)
	{
			uint8_t field_length = p_data[index];
			uint8_t field_type   = p_data[index+1];

			if (field_type == type)
			{
					p_typedata->p_data   = &p_data[index+2];
					p_typedata->data_len = field_length-1;
					return NRF_SUCCESS;
			}
			index += field_length + 1;
	}
	return NRF_ERROR_NOT_FOUND;
}

typedef enum { UNKNOWN_UUID, PHONE_UUID, PHONE_BONDING_UUID, TAG_UUID, TAG_BONDING_UUID } UUID_TYPE;

UUID_TYPE uuid_type(data_t * p_adv_data)
{						
		uint8_t 	phone_uuid[16] = 				{	0x26, 0x53, 0x1d, 0xaa, 0xe9, 0x66, 0x72, 0xb2, 
																				0x22, 0x4a, 0xfa, 0x3a, 0x03, 0x00, 0x75, 0x85 };	
		uint8_t   phone_bonding_uuid[16]= {	0x26, 0x53, 0x1d, 0xaa, 0xe9, 0x66, 0x72, 0xb2, 
																				0x22, 0x4a, 0xfa, 0x3a, 0x04, 0x00, 0x75, 0x85 };
		uint8_t   tag_uuid[16] = 					{	0x26, 0x53, 0x1d, 0xaa, 0xe9, 0x66, 0x72, 0xb2, 
																				0x22, 0x4a, 0xfa, 0x3a, 0x01, 0x00, 0x75, 0x85 };
		uint8_t   tag_bonding_uuid[16] = 	{	0x26, 0x53, 0x1d, 0xaa, 0xe9, 0x66, 0x72, 0xb2, 
																				0x22, 0x4a, 0xfa, 0x3a, 0x02, 0x00, 0x75, 0x85 };
		
		data_t data; 
		uint32_t err_code;
                    
		if((err_code = adv_report_parse(BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_COMPLETE, p_adv_data, &data)) == NRF_SUCCESS)
		{
			if(memcmp(phone_bonding_uuid, data.p_data, 16) == 0)
				return PHONE_BONDING_UUID;
			else if(memcmp(phone_uuid, data.p_data, 16) == 0)
				return PHONE_UUID;
			else
				return UNKNOWN_UUID;
		}
		else if((err_code = adv_report_parse(BLE_GAP_AD_TYPE_MANUFACTURER_SPECIFIC_DATA, p_adv_data, &data)) == NRF_SUCCESS)
		{
			if(memcmp(tag_bonding_uuid, data.p_data + 2, 16) == 0)
				return TAG_BONDING_UUID;
			else if(memcmp(tag_uuid, data.p_data + 2, 16) == 0)
				return TAG_UUID;
			else
				return UNKNOWN_UUID;
		}
		else
			return UNKNOWN_UUID;
}

static void on_ble_evt(ble_evt_t * p_ble_evt)
{
	uint32_t                err_code;
	const ble_gap_evt_t   * p_gap_evt = &p_ble_evt->evt.gap_evt;

	switch (p_ble_evt->header.evt_id)
	{
			case BLE_GAP_EVT_ADV_REPORT:
			{				
					data_t adv_data;           
          adv_data.p_data = (uint8_t *)p_gap_evt->params.adv_report.data;
          adv_data.data_len = p_gap_evt->params.adv_report.dlen;	
				
					UUID_TYPE type = uuid_type(&adv_data);
					
					if(type != UNKNOWN_UUID)
					{						
						err_code = sd_ble_gap_scan_stop();
						APP_ERROR_CHECK(err_code);
						
						app_timer_cnt_get(&m_start_cnt);
						
						err_code = sd_ble_gap_connect((m_scan_param.selective == 1) ? NULL : &p_gap_evt->params.adv_report.peer_addr,
																					&m_scan_param, &m_conn_fast_param);										
						APP_ERROR_CHECK(err_code);
						m_whitelist_temporarily_disabled = false;  
					}							
					
					break;
			}

			case BLE_GAP_EVT_TIMEOUT:
					if(p_gap_evt->params.timeout.src & 
						(BLE_GAP_TIMEOUT_SRC_SCAN | BLE_GAP_TIMEOUT_SRC_CONN | BLE_GAP_TIMEOUT_SRC_SECURITY_REQUEST))
					{                
							scan_start();
					}
					break;
					
			case BLE_GAP_EVT_CONNECTED: 
			{				
					err_code = pm_conn_secure(p_gap_evt->conn_handle, false);
          APP_ERROR_CHECK(err_code);
			}		
					break;
			
			case BLE_GAP_EVT_DISCONNECTED:
					TAG_STATUS_OFF;
					scan_start();
					break;
		 
			case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
					err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle,
																									&p_gap_evt->params.conn_param_update_request.conn_params);
					APP_ERROR_CHECK(err_code);
					break;

			default:
					break;
	}
}

static void on_sys_evt(uint32_t sys_evt)
{
	
	SEGGER_RTT_printf(0, "on_sys_evt: %X\n", sys_evt);
	
	switch (sys_evt)
	{
			case NRF_EVT_FLASH_OPERATION_SUCCESS:           
			case NRF_EVT_FLASH_OPERATION_ERROR:
					if (m_memory_access_in_progress)
					{
							m_memory_access_in_progress = false;
							//scan_start();
					}
					break;
			default:           
					break;
	}
}

static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
	SEGGER_RTT_printf(0, "ble_evt_dispatch: %X\n", p_ble_evt->header.evt_id);

	ble_conn_state_on_ble_evt(p_ble_evt);
	on_ble_evt(p_ble_evt);
	pm_on_ble_evt(p_ble_evt);
}

static void sys_evt_dispatch(uint32_t sys_evt)
{
	SEGGER_RTT_printf(0, "sys_evt_dispatch: %X\n", sys_evt);
		
	fs_sys_event_handler(sys_evt);
	on_sys_evt(sys_evt);
}

#define NRF_CLOCK_SRC	{.source        = NRF_CLOCK_LF_SRC_SYNTH,           \
											 .rc_ctiv       = 0,                                \
											 .rc_temp_ctiv  = 0,                                \
											 .xtal_accuracy = NRF_CLOCK_LF_XTAL_ACCURACY_100_PPM}

static void ble_stack_init(void)
{   
#ifdef DK_BOARD
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
#else
		nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_SRC;
#endif
	
	SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);
	
	ble_enable_params_t ble_enable_params;
	uint32_t err_code = softdevice_enable_get_default_config(CENTRAL_LINK_COUNT, PERIPHERAL_LINK_COUNT, 
																														&ble_enable_params);
	APP_ERROR_CHECK(err_code);    
 
	CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT, PERIPHERAL_LINK_COUNT);    
 
	err_code = softdevice_enable(&ble_enable_params);
	APP_ERROR_CHECK(err_code);
	
	err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
	APP_ERROR_CHECK(err_code);

	err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
	APP_ERROR_CHECK(err_code);
}

static void scan_start(void)
{ 
	uint32_t err_code = sd_ble_gap_scan_start(&m_scan_param);
	APP_ERROR_CHECK(err_code);
}

static void pm_evt_handler(pm_evt_t const * p_evt)
{
    ret_code_t err_code;

    switch(p_evt->evt_id)
    {
        case PM_EVT_BONDED_PEER_CONNECTED:        
            err_code = pm_peer_rank_highest(p_evt->peer_id);
            if(err_code != NRF_ERROR_BUSY)
            {
                APP_ERROR_CHECK(err_code);
            }
						TAG_STATUS_ON;
						break;

        case PM_EVT_CONN_SEC_START:
						break;
				
        case PM_EVT_CONN_SEC_SUCCEEDED:        
            err_code = pm_peer_rank_highest(p_evt->peer_id);
            if(err_code != NRF_ERROR_BUSY)
            {
                APP_ERROR_CHECK(err_code);
            }
						break;

        case PM_EVT_CONN_SEC_FAILED:            
            switch (p_evt->params.conn_sec_failed.error)
            {
                case PM_CONN_SEC_ERROR_PIN_OR_KEY_MISSING:
                    // Rebond if one party has lost its keys.
                    err_code = pm_conn_secure(p_evt->conn_handle, true);
                    if (err_code != NRF_ERROR_INVALID_STATE)
                    {
                        APP_ERROR_CHECK(err_code);
                    }
                    break;

                default:
                    break;
            }
						break;

        case PM_EVT_CONN_SEC_CONFIG_REQ:
						{
							// Reject pairing request from an already bonded peer.
							pm_conn_sec_config_t conn_sec_config = {.allow_repairing = false};
							pm_conn_sec_config_reply(p_evt->conn_handle, &conn_sec_config);
						}
						break;

        case PM_EVT_STORAGE_FULL:
        {
            // Run garbage collection on the flash.
            err_code = fds_gc();
            if (err_code == FDS_ERR_BUSY || err_code == FDS_ERR_NO_SPACE_IN_QUEUES)
            {
                // Retry.
            }
            else
            {
                APP_ERROR_CHECK(err_code);
            }
        }break;

        case PM_EVT_ERROR_UNEXPECTED:            
            APP_ERROR_CHECK(p_evt->params.error_unexpected.error);
            break;

        case PM_EVT_PEER_DATA_UPDATE_SUCCEEDED:
            break;

        case PM_EVT_PEER_DATA_UPDATE_FAILED:            
            APP_ERROR_CHECK_BOOL(false);
            break;

        case PM_EVT_PEER_DELETE_SUCCEEDED:
            break;

        case PM_EVT_PEER_DELETE_FAILED:           
            APP_ERROR_CHECK(p_evt->params.peer_delete_failed.error);
            break;

        case PM_EVT_PEERS_DELETE_SUCCEEDED:
            scan_start();
            break;

        case PM_EVT_PEERS_DELETE_FAILED:            
            APP_ERROR_CHECK(p_evt->params.peers_delete_failed_evt.error);
            break;

        case PM_EVT_LOCAL_DB_CACHE_APPLIED:
            break;

        case PM_EVT_LOCAL_DB_CACHE_APPLY_FAILED:
            // The local database has likely changed, send service changed indications.
            pm_local_database_has_changed();
            break;

        case PM_EVT_SERVICE_CHANGED_IND_SENT:
            break;

        case PM_EVT_SERVICE_CHANGED_IND_CONFIRMED:
            break;

        default:            
            break;
    }
}

static void peer_manager_init(void)
{
    ble_gap_sec_params_t sec_param;
    ret_code_t err_code;

    err_code = pm_init();
    APP_ERROR_CHECK(err_code);

    /*if (erase_bonds)
    {
        err_code = pm_peers_delete();
        APP_ERROR_CHECK(err_code);
    }*/

    memset(&sec_param, 0, sizeof(ble_gap_sec_params_t));

    sec_param.bond              = SEC_PARAM_BOND;
    sec_param.mitm              = SEC_PARAM_MITM;
		sec_param.lesc           		= SEC_PARAM_LESC;
    sec_param.keypress       		= SEC_PARAM_KEYPRESS;
    sec_param.io_caps           = SEC_PARAM_IO_CAPABILITIES;
    sec_param.oob               = SEC_PARAM_OOB;
    sec_param.min_key_size      = SEC_PARAM_MIN_KEY_SIZE;
    sec_param.max_key_size      = SEC_PARAM_MAX_KEY_SIZE;
    sec_param.kdist_own.enc     = 1;
    sec_param.kdist_own.id      = 1;
    sec_param.kdist_peer.enc    = 1;
    sec_param.kdist_peer.id     = 1;

    err_code = pm_sec_params_set(&sec_param);
    APP_ERROR_CHECK(err_code);

    err_code = pm_register(pm_evt_handler);
    APP_ERROR_CHECK(err_code);
}

void ble_db_discovery_evt_handler(ble_db_discovery_evt_t * p_evt)
{
}

static void db_discovery_init(void)
{
	uint32_t err_code = ble_db_discovery_init(ble_db_discovery_evt_handler);
	APP_ERROR_CHECK(err_code);
}

void create_whitelist(bool selective)
{    
		ble_gap_whitelist_t whitelist;
    ble_gap_addr_t    * p_whitelist_addr[BLE_GAP_WHITELIST_ADDR_MAX_COUNT];
    ble_gap_irk_t     * p_whitelist_irk[BLE_GAP_WHITELIST_IRK_MAX_COUNT];
    uint32_t            err_code;
    uint32_t            count;

    // Verify if there is any flash access pending, if yes delay starting scanning until
    // it's complete.
    err_code = fs_queued_op_count_get(&count);
    APP_ERROR_CHECK(err_code);

    if (count != 0)
    {
        m_memory_access_in_progress = true;
        return;
    }

    // Initialize whitelist parameters.
    whitelist.addr_count = BLE_GAP_WHITELIST_ADDR_MAX_COUNT;
    whitelist.irk_count  = 0;
    whitelist.pp_addrs   = p_whitelist_addr;
    whitelist.pp_irks    = p_whitelist_irk;

    // Request creating of whitelist.
    err_code = pm_whitelist_create(NULL, 0, &whitelist);
    APP_ERROR_CHECK(err_code);

    if (((whitelist.addr_count == 0) && (whitelist.irk_count == 0)) ||
        (m_scan_mode != BLE_WHITELIST_SCAN) ||
        (m_whitelist_temporarily_disabled))
    {
        // No devices in whitelist, hence non selective performed.
        m_scan_param.active      = 0;             // Active scanning set.
        m_scan_param.selective   = 0;             // Selective scanning not set.
        m_scan_param.interval    = SCAN_INTERVAL; // Scan interval.
        m_scan_param.window      = SCAN_WINDOW;   // Scan window.
        m_scan_param.p_whitelist = NULL;          // No whitelist provided.
        m_scan_param.timeout     = 0x0000;        // No timeout.
    }
    else
    {
        // Selective scanning based on whitelist first.
        m_scan_param.active      = 0;             // Active scanning set.
        m_scan_param.selective   = 1;             // Selective scanning not set.
        m_scan_param.interval    = SCAN_INTERVAL; // Scan interval.
        m_scan_param.window      = SCAN_WINDOW;   // Scan window.
        m_scan_param.p_whitelist = &whitelist;    // Provide whitelist.
        m_scan_param.timeout     = 0x001E;        // 30 seconds timeout.
    }
}

static void gap_params_init(int8_t tx_power_level)
{	
	char device_name[2 * MAX_STRING_SIZE];
	memset (device_name, 0, 2 * MAX_STRING_SIZE);
	strncat(device_name, hex.name, hex.name_lenght);
	strncat(device_name, hex.serial_number, hex.serial_number_lenght);
			
	ble_gap_conn_sec_mode_t sec_mode;
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);
	uint32_t err_code = sd_ble_gap_device_name_set(&sec_mode, (const uint8_t *)device_name, strlen(device_name));
	APP_ERROR_CHECK(err_code);
	
	ble_opt_t static_pin_option;
	static_pin_option.gap_opt.passkey.p_passkey = (uint8_t *)hex.passkey;
	err_code = sd_ble_opt_set(BLE_GAP_OPT_PASSKEY, &static_pin_option);
	APP_ERROR_CHECK(err_code);
	
	err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_GENERIC_COMPUTER);
	APP_ERROR_CHECK(err_code);

	sd_ble_gap_tx_power_set(tx_power_level);	
}

uint32_t get_bonded_count(void)
{
	return pm_peer_count();
}

int main(void)
{	
	nrf_gpio_cfg_output(RED);
	nrf_gpio_cfg_output(TAG_STATUS);
	LED_ON(RED);
	LED_OFF(RED);
	
	nrf_gpio_cfg_output(TAG_STATUS);
	TAG_STATUS_OFF;

	APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, NULL);
	
	ble_stack_init();	
	ble_conn_state_init();
	peer_manager_init();
	
	bool is_bonded = get_bonded_count() > 0;
	gap_params_init(is_bonded? MAX_TX_POWER_LEVEL : MIN_TX_POWER_LEVEL);
	
	create_whitelist(is_bonded);	
	scan_start();	
	
	while(1)
	{
		uint32_t err_code = sd_app_evt_wait();
		APP_ERROR_CHECK(err_code);
	}
}


